/* Code segments originally written by Nigel Griffiths (nagg@uk.ibm.com)		*/
/* as part of the nmon for Linux tool see web pages:						*/
/*														*/
/* http://www.ibm.com/developerworks/eserver/articles/analyze_aix/     (external) 	*/
/* http://w3.aixncc.uk.ibm.com/tools/pmwiki.php (internal)   				*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
/*#include <curses.h> */
#include <signal.h>
#include <pwd.h>
#include <fcntl.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>
#include <sys/time.h>

#include <netinet/in.h>		/* Internet address structures  */
#include <sys/socket.h>		/* socket interface functions   */
#include <netdb.h>		/* host to IP resolution 	*/

#define	PORT		15050	/* port of "hello world" server */
#define P_CPUINFO	0
#define P_STAT		1
#define P_LPARINFO	2
#define P_MEMINFO   	3
#define P_NUMBER	4 /* one more than the max */
#define CPUMAX 64

char myhost[256];
int debug=0;
int cpus=1;
int loop;
int cpu_idle;
int cpu_user;
int cpu_sys;
int cpu_wait;
double cpu_sum,v1;
int maxloops = -1;
int seconds = -1;
char unixfile[1024];
double *cpu_peak;
int     max_cpus = 1; 
int n,i,pid,smt;
int flash_on=0;
int			rc;       /* system calls return value storage  */
int			s;        /* socket descriptor                  */
int			cs;       /* new connection's socket descriptor */
struct sockaddr_in	sa;       /* Internet address struct            */
struct sockaddr_in	csa;      /* client's address struct            */
int               	size_csa; /* size of client's address struct    */
char		mystr[256];
char        buffer[256];
long		tloc,next_minute;
unsigned long 	last_time_base,last_purr;
double elapsed;

struct info_stat{
	long timebase;
};

struct cpu_stat {
	long user;
	long sys;
	long wait; /* sometimes nice */
	long idle;
};

struct mem_stat {
	long memtotal;
	long memfree;
	long memshared;
	long buffers;
	long cached;
	long swapcached;
	long active;
	long inactive;
	long hightotal;
	long highfree;
	long lowtotal;
	long lowfree;
	long swaptotal;
	long swapfree;
	long bigfree;
};

struct lpar_stat {
        char version[8]; 
        char serial[8];
	long partition_id;
	long DisWheRotPer;
	long MinEntCap;
	long MinEntCapPerVP;
	long MinMem;
	long MinProcs;
	long partition_max_entitled_capacity;
	long system_potential_processors;
	long DesEntCap;
	long DesMem;
	long DesProcs;
	long DesVarCapWt;
	long DedDonMode;
	long partition_entitled_capacity;
	long system_active_processors;
	long capacity_weight;
	long capped;
        long entitled_memory;
        long backing_memory;
	unsigned long purr;
	long partition_active_processors;
	long partition_potential_processors;
	long shared_processor_mode;
};




struct data {
	struct cpu_stat cpu_total;
	struct cpu_stat cpuN[CPUMAX];
	struct mem_stat mem;
	struct lpar_stat lpar;
	struct info_stat info;
#ifdef PARTITIONS
	struct part_stat parts[PARTMAX];
#endif /*PARTITIONS*/

	struct timeval tv;
	double time;

} database[2], *p, *q;




int reread =0;
struct {
	FILE *fp;
	char *filename;
	char buf[1024*4];
	char *line[64];
	int lines;
} proc[P_NUMBER];


double  doubletime(void)
{

        gettimeofday(&p->tv, 0);
        return((double)p->tv.tv_sec + p->tv.tv_usec * 1.0e-6);
}


proc_init()
{
int i;
	for(i=0;i<P_NUMBER;i++)
		proc[i].fp = 0;

	proc[P_CPUINFO].filename = "/proc/cpuinfo";
	proc[P_STAT].filename    = "/proc/stat";
	proc[P_LPARINFO].filename = "/proc/ppc64/lparcfg";
	proc[P_MEMINFO].filename  = "/proc/meminfo";
}

proc_read(int num)
{
int i;
int size;
int found;

	if(proc[num].fp == 0) {
		if( (proc[num].fp = fopen(proc[num].filename,"r")) == NULL) {
			perror("failed to open");
			printf("filename=%s\n",proc[num].filename);
		}
	}
	rewind(proc[num].fp);
	size = fread(proc[num].buf, 1, 1024*4-1, proc[num].fp);
	proc[num].buf[size]=0;
	proc[num].lines=0;
	proc[num].line[0]=&proc[num].buf[0];
	for(i=0;i<size;i++) {
		if(proc[num].buf[i] == '\t') 
			proc[num].buf[i]= ' '; 
		if(proc[num].buf[i] == '\n') {
			proc[num].lines++;
			proc[num].buf[i] = 0;
			proc[num].line[proc[num].lines] = &proc[num].buf[i+1];
		}
	}
	if(reread) {
		fclose( proc[num].fp);
		proc[num].fp = 0;
	}
}

void switcher(void)
{
	static int	which = 1;

	if (which) {
		p = &database[0];
		q = &database[1];
		which = 0;
	} else {
		p = &database[1];
		q = &database[0];
		which = 1;
	}
	if(flash_on)
		flash_on = 0;
	else
		flash_on = 1;
}

proc_mem()
{
	sscanf(&proc[P_MEMINFO].line[0][11],  "%ld", &p->mem.memtotal);
	sscanf(&proc[P_MEMINFO].line[1][11],  "%ld", &p->mem.memfree);
/*
	sscanf(&proc[P_MEMINFO].line[2][11],  "%ld", &p->mem.memshared);
	sscanf(&proc[P_MEMINFO].line[2][11],  "%ld", &p->mem.buffers);
	sscanf(&proc[P_MEMINFO].line[3][11],  "%ld", &p->mem.cached);
	sscanf(&proc[P_MEMINFO].line[4][11],  "%ld", &p->mem.swapcached);
	sscanf(&proc[P_MEMINFO].line[5][11],  "%ld", &p->mem.active);
	sscanf(&proc[P_MEMINFO].line[6][11], "%ld", &p->mem.inactive);
	sscanf(&proc[P_MEMINFO].line[7][11], "%ld", &p->mem.hightotal);
	sscanf(&proc[P_MEMINFO].line[8][11], "%ld", &p->mem.highfree);
	sscanf(&proc[P_MEMINFO].line[9][11], "%ld", &p->mem.lowtotal);
	sscanf(&proc[P_MEMINFO].line[10][11], "%ld", &p->mem.lowfree);
	sscanf(&proc[P_MEMINFO].line[11][11], "%ld", &p->mem.swaptotal);
	sscanf(&proc[P_MEMINFO].line[12][11], "%ld", &p->mem.swapfree);
*/		
}

proc_cpu()
{
int i;
	sscanf(&proc[P_STAT].line[0][5], "%ld %ld %ld %ld", 
			&p->cpu_total.user,
			&p->cpu_total.wait,
			&p->cpu_total.sys,
			&p->cpu_total.idle);
	for(i=0;i<cpus;i++ )
		sscanf(&proc[P_STAT].line[i+1][5], "%ld %ld %ld %ld", 
			&p->cpuN[i].user,
			&p->cpuN[i].wait,
			&p->cpuN[i].sys,
			&p->cpuN[i].idle);

}

proc_cpuinfo()
{
int i;
i=proc[P_CPUINFO].lines-4;
if (debug==1) printf("icpuinfo line %s \n",proc[P_CPUINFO].line[i]);
sscanf(&proc[P_CPUINFO].line[i][11], "%ld",&p->info.timebase);
    if (debug==1) printf("timebase from file=%.d \n",p->info.timebase); 
}


proc_lpar()
{
int i;

	sscanf(&proc[P_LPARINFO].line[0][8],  "%s" ,&p->lpar.version);	
	sscanf(&proc[P_LPARINFO].line[1][21],  "%s" ,&p->lpar.serial);	

	sscanf(&proc[P_LPARINFO].line[3][13],  "%ld" ,&p->lpar.partition_id);	
	sscanf(&proc[P_LPARINFO].line[19][28],  "%ld", &p->lpar.partition_entitled_capacity);
	sscanf(&proc[P_LPARINFO].line[21][25],  "%ld", &p->lpar.system_active_processors);
	sscanf(&proc[P_LPARINFO].line[27][16],  "%ld", &p->lpar.capacity_weight);
	sscanf(&proc[P_LPARINFO].line[28][7],  "%ld", &p->lpar.capped);

/*	sscanf(&proc[P_LPARINFO].line[34][5],  "%ld", &p->lpar.purr) */
      p->lpar.purr=strtoul(&proc[P_LPARINFO].line[39][5],NULL,12);
    if (debug==1) printf("purr from file=%.3f \n",p->lpar.purr); 

	sscanf(&proc[P_LPARINFO].line[40][28],  "%ld", &p->lpar.partition_active_processors);
	sscanf(&proc[P_LPARINFO].line[42][22],  "%ld", &p->lpar.shared_processor_mode);
}

void dostuff (int sock)
{	    
unsigned long purr,purr1,time_base;
long entitled_purr, unused_purr;
unsigned long delta_purr, delta_tb;
double new_ent;
double phy_proc_consumed = 0, per_ent;	

	
	for(loop=1; ; loop++) {
				bzero(buffer,256);

				proc_read(P_STAT);
				proc_cpu();
				proc_read(P_MEMINFO);
				proc_mem();
				proc_read(P_CPUINFO);
				proc_cpuinfo();
				proc_read(P_LPARINFO);
				proc_lpar(); 
                                
				cpu_user = p->cpu_total.user - q->cpu_total.user; 
				cpu_sys  = p->cpu_total.sys  - q->cpu_total.sys; 
				cpu_wait = p->cpu_total.wait - q->cpu_total.wait; 
				cpu_idle = p->cpu_total.idle - q->cpu_total.idle; 
				cpu_sum = cpu_idle + cpu_user + cpu_sys + cpu_wait;
				
				
    new_ent = (double)p->lpar.partition_entitled_capacity / 100.0 ;
    /* compute values to be displayed */
    /* 1- processor utilization ressource register */
    purr1 = cpu_sum;
    purr=p->lpar.purr;
    delta_purr = purr - last_purr;

	p->time = doubletime();
	elapsed = p->time - q->time;

    /* 2 - time elapsed */
    time_base = p->info.timebase;
    /*   delta_tb = time_base - last_time_base; */
    delta_tb = time_base;
    
    /* gives the physical pocessor consumed */
    phy_proc_consumed = (double)delta_purr / (double)time_base;
 
   phy_proc_consumed=phy_proc_consumed/elapsed;

    if (debug==1) printf("phys=%.3f \n",phy_proc_consumed); 

    if (phy_proc_consumed < 0.0) 
	phy_proc_consumed=0.0;
    if (phy_proc_consumed > (double)p->lpar.partition_active_processors)
    	phy_proc_consumed = (double)p->lpar.partition_active_processors;
 
    if ((int)p->lpar.capacity_weight == 0) {
     if (phy_proc_consumed > (double)p->lpar.partition_entitled_capacity/100.0)
	phy_proc_consumed = (double)p->lpar.partition_entitled_capacity/100.0;
    }
    
    if (debug==1)  printf("purr=%ld purr1=%ld deltap= %ld time_base=%ld delta_tb=%ld phys=%.2f elapsed=%.2f vp=%ld \n",
    purr,purr1,delta_purr,time_base,delta_tb,phy_proc_consumed,elapsed,p->lpar.partition_active_processors);
    
    per_ent = (double)((phy_proc_consumed / new_ent) * 100);
    rc=gethostname(myhost,256);
				
				
sprintf(mystr,"PLS61,%s,%.1f,%.1f,%.1f,%.1f,%.2f,%.1f,%.1f,%.1f,%u,%s,%u,%u,%u,%s,%s,%u,%.1f,%.1f,0,0,0,0,0,%s,0,0,",  
				    p->lpar.shared_processor_mode ? "Shared" : "Dedicated",
				    (double)cpu_user / (double)cpu_sum * 100.0,
				    (double)cpu_sys  / (double)cpu_sum * 100.0,
				    (double)cpu_wait / (double)cpu_sum * 100.0,
				    (double)cpu_idle / (double)cpu_sum * 100.0,
				    (double)phy_proc_consumed,
        			per_ent,
                		(double)phy_proc_consumed,
        			per_ent,
				p->lpar.system_active_processors,
				p->lpar.capped ? "Capped" : "Uncapped",	
				p->lpar.capacity_weight,
	        		p->lpar.partition_active_processors,
				p->lpar.partition_entitled_capacity,
				smt ? "SMT On" : "SMT Off",
                                myhost,
				p->lpar.partition_id,    
				(double)p->mem.memtotal/1024.0,
				(double)p->mem.memfree/1024.0,
				p->lpar.serial);
				
/*printf("%ld \n",p->mem.memtotal); 				
printf("%s \n",mystr); 		*/		
	/* ok, we got data ... do the job... */

					n=write(sock,mystr, sizeof(mystr));
						if (n<0) error("error reading from socket");
				      if (debug) printf("wrote %s %d bytes - \n",mystr,n);
					n=read(sock,buffer,sizeof(buffer));
        				if (n<0) error("error reading from socket");
					if (debug==1) printf("%s \n",buffer);

					if (strncmp(buffer,"exit",4)==0)
					{
						return;
					}

/*					if (debug==1) printf("%s \n",buffer); */

				switcher();
				    
    				last_purr = purr;
    				last_time_base = time_base;
				reread=1;
	}				
}




main()
{
    signal(SIGCHLD,SIG_IGN);
    /* initiate machine's Internet address structure */
    /* first clear out the struct, to avoid garbage  */
    memset(&sa, 0, sizeof(sa));
    /* Using Internet address family */
    sa.sin_family = AF_INET;
    /* copy port number in network byte order */
    sa.sin_port = htons(PORT);
    /* we will accept cnnections coming through any IP	*/
    /* address that belongs to our host, using the	*/
    /* INADDR_ANY wild-card.				*/
    sa.sin_addr.s_addr = INADDR_ANY;

    /* allocate a free socket                 */
    /* Internet address family, Stream socket */
    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	perror("socket: allocation failed");
    }

    /* bind the socket to the newly formed address */
    rc = bind(s, (struct sockaddr *)&sa, sizeof(sa));

    /* check there was no error */
    if (rc) {
	perror("bind");
    }

    /* ask the system to listen for incoming connections	*/
    /* to the address we just bound. specify that up to		*/
    /* 5 pending connection requests will be queued by the	*/
    /* system, if we are not directly awaiting them using	*/
    /* the accept() system cal, when they arrive.		*/
    rc = listen(s, 5);

    /* check there was no error */
    if (rc) {
	perror("listen");
    }
    /* remember size for later usage */
    size_csa = sizeof(csa);
    /* enter an accept-write-close infinite loop */





	proc_init();
	/* Set parameters if not set by above */
	if (maxloops == -1)
		maxloops = 9999999;
	if (seconds  == -1)
		seconds = 2;
	/* To get the pointers setup */
	switcher();

	/* Request the Kernel addresses - if this fails stop now */
	readlink("/unix", unixfile, 1024);

	/* Initialise the time stamps for the first loop */
	p->time = doubletime();
	q->time = doubletime();

	cpu_peak = malloc(sizeof(double) * 128); /* MAGIC */
	for(i=0;i<(max_cpus+1);i++)
		cpu_peak[i]=0.0;

	switcher();
		
	proc_read(P_STAT);
        for(i=1;i<proc[P_STAT].lines;i++) {
                if(strncmp("cpu",proc[P_STAT].line[i],3) == 0)
                        max_cpus = cpus=i;
                else
                        break;
        }	
/*	printf("cpus= %u \n",max_cpus); */
        smt=0;
	if (max_cpus>p->lpar.partition_active_processors)
		smt=1;	

      proc_read(P_LPARINFO);
	proc_lpar(); 
	v1=atof(p->lpar.version);			
	rc=strncmp(p->lpar.version,"1.9",3);
	if (v1 != 1.8) {
                printf("/proc/ppc64/lparcfg is not version 1.8 \n");
                printf("You must use the right version of the linuxagent\n");
                return;   
		kill(getpid(),SIGKILL);
		};
		
    	while (1) {
       	/* the accept() system call will wait for a	*/
       	/* connection, and when one is established, a	*/
       	/* new socket will be created to form it, and	*/
       	/* the csa variable will hold the address	*/
       	/* of the Client that just connected to us.	*/
       	/* the old socket, s, will still be available	*/
       	/* for future accept() statements.		*/
       	
   
	cs = accept(s, (struct sockaddr *)&csa, &size_csa);

       	/* check for errors. if any, enter accept mode again */
       	if (cs < 0)
    	    continue;
	    
	pid=fork();    
	if (pid <0)
		error("ERROR on fork");
	if (pid == 0) 	{
		close(s);
		dostuff(cs);
		kill(getpid(),SIGKILL);
		}
	else close(cs);

	}   /* end while */
}	

