#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <libperfstat.h>
#include <netinet/in.h>		/* Internet address structures  */
#include <sys/socket.h>		/* socket interface functions   */
#include <netdb.h>		/* host to IP resolution 	*/
#include <sys/utsname.h>	/* for serial number */
#include <sys/dr.h>
#include <sys/vminfo.h>

#define	PORT		15050	/* port for "gmon" agent */

static int disp_util_header = 1;

/* Data structures to hold old values */
static u_longlong_t last_purr, last_time_base;
static u_longlong_t last_vcpu_user, last_vcpu_sys, last_vcpu_idle, last_vcpu_wait;
static u_longlong_t last_phint = 0, last_vcsw = 0, last_pit = 0;
static u_longlong_t last_co = 0 ,last_ci = 0, last_hpi = 0;

static u_longlong_t last_xfers,last_wblks,last_rblks;
static u_longlong_t last_ipackets,last_ibytes,last_opackets,last_obytes;

static u_longlong_t last_pgspins,last_pgspouts,last_rq,last_sq;
static u_longlong_t last_scans,last_cycles,last_pgsteals;

struct dr_info drInfo;

/*void display_lpar_util(void);*/
int			rc,loop,debug;       
int			s;        /* socket descriptor                  */
int			cs;       /* new connection's socket descriptor */
struct sockaddr_in	sa;       /* Internet address struct            */
struct sockaddr_in	csa;      /* client's address struct            */
int               	size_csa; /* size of client's address struct    */
char		mystr[768];
char            buffer[768];
char            vers[8];
long		tloc,next_minute;
int 		n,i,pid;
int 		reconfig,drtype,drcmd;
size_t		bufsize;

struct utsname uts;

void catch_reconfig(int sig_num)
{
    /* re-set the signal handler again to catch_int, for next time */
    signal(SIGRECONFIG,catch_reconfig); 

/*assume it's not PM */
    drtype=4;
/*get drtype */
    rc = dr_reconfig(DR_QUERY, &drInfo);

/*set drtype */

    if (drInfo.migrate) {
        if (drInfo.check) {
		drtype=1;
	} else if (drInfo.pre) {
		drtype=2;
	} else if (drInfo.post) {
                drtype=3;
        }
    } /* end drinfo.migrate */
    reconfig=1;
/* ack DR event */
    rc = dr_reconfig(DR_RECONFIG_DONE, &drInfo);

}


/********************************************************
* This function display_lpar_util(), displays the utilization
* metrics for the LPAR along with the LPAR details.
*********************************************************/
void display_lpar_util()
{
    u_longlong_t dlt_vcpu_user, dlt_vcpu_sys, dlt_vcpu_idle, dlt_vcpu_wait;
    u_longlong_t dlt_pgspins,dlt_pgspouts,frame_mem;
    u_longlong_t dlt_scans,dlt_cycles,dlt_pgsteals;
    u_longlong_t vcptimes;
    u_longlong_t purr, time_base;
    u_longlong_t entitled_purr, unused_purr;
    u_longlong_t delta_purr, delta_tb;
    u_longlong_t vcsw;
    u_longlong_t dlt_ci,dlt_co,ci,co;
    double new_ent;
    double phy_proc_consumed = 0, per_ent;
    int i,j, rc,vsize;
    perfstat_partition_total_t partitionstats;
    perfstat_cpu_total_t cpustats;
    perfstat_memory_total_t minfo;
    perfstat_disk_total_t diskstats;
    perfstat_netinterface_total_t netstats;

    
    u_longlong_t xfers,wblks,rblks,rq,sq,hpi;
    u_longlong_t ipackets,ibytes,opackets,obytes;	

    lpar_info_format2_t lparinfo2;
    struct vminfo vmi;

    /* get the metrics */

    uname(&uts);
    uts.machine[strlen(uts.machine)-4] = 0;

    rc = perfstat_partition_total(NULL, &partitionstats,
                                  sizeof(perfstat_partition_total_t), 1);
    if (rc != 1) {
        perror("perfstat_partition_total");
        exit(-1);
    }
    perfstat_cpu_total(NULL,&cpustats,sizeof(perfstat_cpu_total_t),1);
    perfstat_memory_total(NULL,&minfo,sizeof(perfstat_memory_total_t),1);
    perfstat_disk_total(NULL,&diskstats,sizeof(perfstat_disk_total_t),1);
    perfstat_netinterface_total(NULL,&netstats,sizeof(perfstat_netinterface_total_t),1);
    lparinfo2.lpar_flags=LPAR_INFO2_EXTENDED;
  rc=lpar_get_info(2,&lparinfo2,sizeof(lpar_info_format2_t)); 
/*     printf("lpar_get_info  %u flags: %u \n",rc,lparinfo2.lpar_flags); */
/* Print the header for utilization metrics */
    if (vmgetinfo((void *)&vmi,VMINFO,sizeof(vmi)) != 0)
	{
	perror("vmgetinfo failed");
	}
    ci=vmi.cmem_ncomp_ops;
    co=vmi.cmem_ndecomp_ops;
    dlt_ci=ci-last_ci;
    dlt_co=co-last_co;
    new_ent = (double)partitionstats.entitled_proc_capacity / 100.0 ;


    /* compute values to be displayed */
    /* 1- processor utilization ressource register */
    purr = partitionstats.puser + partitionstats.psys +
           partitionstats.pidle + partitionstats.pwait;
    delta_purr = purr - last_purr;
    /* 2 - time elapsed */
    time_base = partitionstats.timebase_last;
    delta_tb = time_base - last_time_base;
    /* gives the physical pocessor consumed */
    phy_proc_consumed = (double)delta_purr / (double)delta_tb;

    /* 3 - user/sys/idle/wait use computation */
    dlt_vcpu_user = partitionstats.puser - last_vcpu_user;
    last_vcpu_user = partitionstats.puser;
    dlt_vcpu_sys = partitionstats.psys - last_vcpu_sys;
    last_vcpu_sys = partitionstats.psys;
    dlt_vcpu_idle = partitionstats.pidle - last_vcpu_idle;
    last_vcpu_idle = partitionstats.pidle;
    dlt_vcpu_wait = partitionstats.pwait - last_vcpu_wait;
    last_vcpu_wait = partitionstats.pwait;


/* disk io */

	xfers=diskstats.xfers - last_xfers;
	last_xfers=diskstats.xfers;

	wblks=diskstats.wblks - last_wblks;
	last_wblks=diskstats.wblks;

	rblks=diskstats.rblks - last_rblks;
	last_rblks=diskstats.rblks;
    

/* net io */

	ipackets=netstats.ipackets - last_ipackets;
	last_ipackets=netstats.ipackets;

	ibytes=netstats.ibytes - last_ibytes;
	last_ibytes=netstats.ibytes;

	opackets=netstats.opackets - last_opackets;
	last_opackets=netstats.opackets;

	obytes=netstats.obytes - last_obytes;
	last_obytes=netstats.obytes;


    /* paging */

    dlt_pgspins=minfo.pgspins-last_pgspins;
    last_pgspins=minfo.pgspins;

    dlt_pgspouts=minfo.pgspouts-last_pgspouts;
    last_pgspouts=minfo.pgspouts;

    dlt_scans=minfo.scans-last_scans;
    last_scans=minfo.scans;

    dlt_cycles=minfo.cycles-last_cycles;
    last_cycles=minfo.cycles;

    dlt_pgsteals=minfo.pgsteals-last_pgsteals;
    last_pgsteals=minfo.pgsteals;

/* runque*/

    rq=cpustats.runque-last_rq;
    last_rq=cpustats.runque;

    sq=cpustats.swpque-last_sq;
    last_sq=cpustats.swpque;

    hpi=partitionstats.hpi-last_hpi;
    last_hpi=partitionstats.hpi;

    vcptimes = dlt_vcpu_user + dlt_vcpu_sys + dlt_vcpu_idle + dlt_vcpu_wait;

    if ( partitionstats.type.b.shared_enabled ||  partitionstats.type.b.donate_enabled )
    { /* recompute some values to compensate shared mode errors */
        entitled_purr = delta_tb * new_ent;
        if (entitled_purr < vcptimes) {entitled_purr = vcptimes;}
        unused_purr = entitled_purr - vcptimes;
        dlt_vcpu_wait += unused_purr * ((double)
                         partitionstats.pwait /
                         (double)(partitionstats.pwait +
                         partitionstats.pidle));
        dlt_vcpu_idle += unused_purr * ((double)
                         partitionstats.pidle /
                         (double)(partitionstats.pwait +
                         partitionstats.pidle));
        vcptimes = entitled_purr;
 /*       dlt_vcpu_idle+=dlt_vcpu_wait;
        dlt_vcpu_wait-=dlt_vcpu_wait; */

    }
    else
    {
    phy_proc_consumed=(double)((100.0-(double)dlt_vcpu_idle*100.0/(double)vcptimes)/100.0)*partitionstats.online_cpus;
    }


    per_ent = (double)((phy_proc_consumed / new_ent) * 100);

/* formatting is important - ushort is %d size64_t is %lld */ 

	sprintf(mystr,
	"%s,%s,%.1f,%.1f,%.1f,%.1f,%.2f,%.1f,%.1f,%.1f,%u,%s,%u,%u,%u,SMT=%u,%s,%u,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%s,%u,%u,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%d,%lld,%llu,%u,%u,%.2f,%.2f,%llu,%llu,%llu,%llu,%llu,%llu,",
		vers,
		partitionstats.type.b.shared_enabled ? "Shared" : "Dedicated",	
		(double)dlt_vcpu_user*100.0/(double)vcptimes,
    		(double)dlt_vcpu_sys*100.0/(double)vcptimes,
    		(double)dlt_vcpu_wait*100.0/(double)vcptimes,
                (double)dlt_vcpu_idle*100.0/(double)vcptimes,
                (double)phy_proc_consumed,
        	per_ent,
                ((double)dlt_vcpu_user*100.0/(double)vcptimes)+
                         ((double)dlt_vcpu_sys*100.0/(double)vcptimes),
                (double)(partitionstats.pool_idle_time - last_pit) /
                         (double)delta_tb,
	        partitionstats.online_phys_cpus_sys,
		partitionstats.type.b.capped ? "Capped" : "Uncapped",	
		partitionstats.var_proc_capacity_weight,
		partitionstats.online_cpus,
		partitionstats.entitled_proc_capacity,
		partitionstats.smt_thrds,
		partitionstats.name,
		partitionstats.lpar_id,
		minfo.real_total*4096/1024/1024,
		minfo.real_free*4096/1024/1024,
		dlt_pgspins,
		dlt_pgspouts,
		dlt_scans,
   		dlt_cycles,
   		dlt_pgsteals,
		&uts.machine[2],
		reconfig,
                drtype,
		partitionstats.online_memory,
		minfo.real_inuse*4096/1024/1024,
		minfo.real_system*4096/1024/1024,
		minfo.real_user*4096/1024/1024,
		minfo.real_process*4096/1024/1024,
		minfo.numperm*4096/1024/1024,
		minfo.pgsp_total*4096/1024/1024,
		minfo.pgsp_free*4096/1024/1024,
		minfo.pgsp_rsvd*4096/1024/1024,
		minfo.virt_total*4096/1024/1024,
		minfo.virt_active*4096/1024/1024,
		partitionstats.max_memory,
		partitionstats.min_memory,
		xfers,
		wblks,
		rblks,
		ipackets,
	        ibytes,
		opackets,
		obytes,
		rq,
		hpi,
		cpustats.processorHZ,		
		minfo.pmem/1024/1024,
		lparinfo2.vrm_pool_id,
		lparinfo2.vrm_pool_physmem/1024/1024,
                lparinfo2.ame_online_memory,
                lparinfo2.ame_factor,
       		lparinfo2.ame_type,
 		(double)vmi.ame_factor_tgt/100.0,
 		(double)vmi.ame_factor_actual/100.0,
		vmi.ame_deficit_size/1024/1024,
 		vmi.cmem_cpool_size/1024/1024,
 		vmi.cmem_cpool_free/1024/1024,
 		vmi.cmem_ucpool_size/1024/1024,
 		dlt_ci,
 		dlt_co
		);

    /* Store the current values for the next iteration */
    last_vcsw = vcsw;
    last_phint = partitionstats.phantintrs;
    last_pit = partitionstats.pool_idle_time;
    last_purr = purr;
    last_time_base = time_base;
    last_ci=ci;
    last_co=co;


    return;
}


void dostuff (int sock)
{	    
	for(loop=1; ; loop++) {
	bzero(buffer,512);
			
	display_lpar_util();				
	/* ok, we got a new connection. do the job... */
	n=write(sock,mystr, sizeof(mystr));
	if (n<0) perror("error reading from socket");
	        if (debug==1) printf("wrote %s %d bytes - ",mystr,n);
		n=read(sock,buffer,sizeof(buffer));
       		if (n<0) perror("error reading from socket");
		if (debug == 1) printf("%s \n",buffer);
 		if (strncmp(buffer,"exit",5)==0)
		{
		break;
		}
		if (strncmp(buffer,"ack2",4)==0)
		{
	        reconfig=0;
		drtype=0;	
		}



	}				
}


/* main loop */
int main(int argc, char* argv[])
{
    int aixver,aixtl,aixsp;
    FILE *fp;
    char str[256];
    
    debug=0;
    drtype=0;
    reconfig=0;
/* check AIX versions */
    sprintf(vers,"%s","AIX613");

    fp=popen("oslevel -s","r");
    if (fp == NULL) {
        perror("popen oslevel");
        printf("popen error \n!");
        }
    else
    {
    if ( fgets(str,256,fp) != NULL ) {
         sscanf(str,"%d-%d-%d",&aixver,&aixtl,&aixsp);
         if ( aixver < 5300 ) {
         printf("aix version not supported \n");
         exit(0);
	 }
         if ( aixver == 5300 ) {
            if (aixtl < 11 ) {
		printf("AIX 5.3 must be at TL11 or above \n");
    	        exit(0);
	        }
	 }
         if ( aixver == 6100 ) {
		if (aixtl >= 4) {
			sprintf(vers,"%s","AIX614");
		}	
		else
		{
		printf("AIX 6.1 must be at TL4 or above \n");
    	        exit(0);
		}
	 }
	 if (aixver > 6100 ) {
			sprintf(vers,"%s","AIX614");
		}
         }
        else
        {
        printf("warning no aix vers \n");
        }

    }
    pclose(fp);


    signal(SIGCHLD,SIG_IGN);
    signal(SIGRECONFIG,catch_reconfig);    

    /* initiate machine's Internet address structure */
    /* first clear out the struct, to avoid garbage  */
    memset(&sa, 0, sizeof(sa));
    /* Using Internet address family */
    sa.sin_family = AF_INET;
    /* copy port number in network byte order */
    sa.sin_port = htons(PORT);
    /* we will accept cnnections coming through any IP	*/
    /* address that belongs to our host, using the	*/
    /* INADDR_ANY wild-card.				*/
    sa.sin_addr.s_addr = INADDR_ANY;

    /* allocate a free socket                 */
    /* Internet address family, Stream socket */
    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	perror("socket: allocation failed");
    }

    /* bind the socket to the newly formed address */
    rc = bind(s, (struct sockaddr *)&sa, sizeof(sa));

    /* check there was no error */
    if (rc) {
	perror("bind");
    }
    setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(void *)&sa,sizeof(sa));

    /* ask the system to listen for incoming connections	*/
    /* to the address we just bound. specify that up to		*/
    /* 5 pending connection requests will be queued by the	*/
    /* system, if we are not directly awaiting them using	*/
    /* the accept() system cal, when they arrive.		*/
    rc = listen(s, 5);

    /* check there was no error */
    if (rc) {
	perror("listen");
    }
    /* remember size for later usage */
    size_csa = sizeof(csa);
    /* enter an accept-write-close infinite loop */
    while (1) {
   
	cs = accept(s, (struct sockaddr *)&csa, &size_csa);
       	/* check for errors. if any, enter accept mode again */
       	if (cs < 0)
    	    continue;
	pid=fork();    
	if (pid <0)
		perror("ERROR on fork");
	if (pid == 0) 	{
		close(s);
                dostuff(cs);
		kill(getpid(),SIGKILL);;    
	}
	else close(cs);
    }
    return(0);
}




